/*\
 * hiscore.c - code to handle reading from/writing to hiscore table
 * asteroid.exe: a version of Asteroids by Charles Lindsay
 * Copyright (C) 2003, 2005 Charles Lindsay
 *
 * Asteroids is originally copyright (C) 1979 Atari, Inc.
 *
 * This version of Asteroids is released under the GNU GPL and as
 * such, comes with absolutely no warranty of any kind, whatsoever.
 * See asteroid-license.txt for more information.
 *
 * webpage: http://chaoslizard.sourceforge.net/asteroid/
 * contact: chaoslizard@gamebox.net
 *
 * date        author   comments
 * ----        ------   --------
 * 2003-12-21  Charles  initial implementation
 * 2003-12-22  Charles  more functionality
 * 2003-12-23  Charles  added gethiscorename, checksums
 * 2005-02-16  Charles  removed utter dependence on Win32
 * 2005-03-09  Charles  used g_szPlayerName from main.c to do GetHiScoreName when not in Win32
 * 2005-03-12  Charles  changed hiscore file name and fixed potential segfault in linux version
 *                      also, added openhiscoretable and validatehiscorename
 *
\*/


#ifdef _WIN32
#include <windows.h>
#include "resource.h"
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <stdio.h>
#include <ctype.h>

#include "main.h"
#include "hiscore.h"
#include "render.h"


#ifdef _WIN32
#define HS_FILE "hiscores"
#else
#define HS_FILE ".asteroid_hiscores"
#endif


void Crypt(char *, unsigned int); //encrypts a buffer
char CheckSum(const char *, unsigned int); //gets the checksum of buffer (minus last byte)
#ifdef _WIN32
INT_PTR CALLBACK HiScoreDialog(HWND, UINT, WPARAM, LPARAM); //dialog box callback

char * g_name = NULL; //used by HiScoreDialog--destination name buffer
#else
extern char g_szPlayerName[HS_NAME_LENGTH]; //from main.c, the name as it was entered at the prompt
#endif


//reads the table from the hiscore file
void ReadHiScoreTable(HISCORE_TABLE * h)
{
   FILE * f = NULL;
   char * contents = NULL;
   unsigned int len;
   unsigned char entries;

   if(h)
   {
      do
      {
         memset(h, 0, sizeof(HISCORE_TABLE)); //zero the table

         f = OpenHiScoreTable("rb"); //open the file
         if(!f)
            break; //if it can't open, break

         fseek(f, 0, SEEK_END);
         len = ftell(f); //get length of file
         if(len <= 2 || len > (2 + HS_ENTRIES*sizeof(HST_ENTRY)) || ((len - 2) % sizeof(HST_ENTRY)))
            break; //if length is invalid, break

         contents = ast_alloc(len); //alloc space for the file
         fseek(f, 0, SEEK_SET);
         fread(contents, 1, len, f); //read file in

         Crypt(contents, len); //decrypt file

         entries = contents[0]; //get how many entries
         if(entries > HS_ENTRIES || len < 2 + entries*sizeof(HST_ENTRY))
            break; //if invalid entries, break (note that len can be greater than number of entries allows for)

         if(contents[1 + entries*sizeof(HST_ENTRY)] != CheckSum(contents, 2 + entries*sizeof(HST_ENTRY)))
            break; //if checksum fails, break

         memcpy(h->entries, contents + 1, entries*sizeof(HST_ENTRY)); //copy the table into memory

      } while(0);

      if(f)
         fclose(f);
      if(contents)
         free(contents);
   }
}

//returns 1 if the score belongs in the table, 0 if it didn't make it
int IsScoreHi(const HISCORE_TABLE * h, unsigned int s)
{
   int i;
   for(i = 0; i < HS_ENTRIES; i++)
   {
      if(s > h->entries[i].score)
         return 1;
   }
   return 0;
}

//retrieves the player's name for a hi score
int GetHiScoreName(char * buf)
{
#ifdef _WIN32
   g_name = buf;
   if(DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_HISCORE), NULL, HiScoreDialog))
   {
      ValidateHiScoreName(buf);
      return 1;
   }
   return 0;
#else
   strcpy(buf, g_szPlayerName); //g_szPlayerName is already validated
   return 1;
#endif
}

//removes unprintable characters from the string
void ValidateHiScoreName(char * name)
{
   char * t;
   char * u;

   for(t = name; *t; ++t) //look at each character in string
   {
      if(!isprint(*t)) //if it's not printable
      {
         for(u = t; *u; ++u) //copy everything down one
            *u = *(u + 1);
      }
   }
}

//adds a score to the hiscore table and writes it to the file
void AddHiScore(HISCORE_TABLE * h, const char * n, unsigned int s, unsigned int l)
{
   FILE * f = NULL;
   char * buf = NULL;
   unsigned int len;
   unsigned char entries;
   int i;

   if(h && n && s && l)
   {
      for(i = 0; i < HS_ENTRIES; i++)
      {
         if(s > h->entries[i].score) //when we find the first entry to put the new score before
         {
            //move everything else down
            memmove(&h->entries[i + 1], &h->entries[i], sizeof(HST_ENTRY) * (HS_ENTRIES - 1 - i));

            //add the entry here
            h->entries[i].score = s;
            strncpy(h->entries[i].name, n, HS_NAME_LENGTH);
            h->entries[i].level = l;
            break;
         }
      }

      do
      {
         for(i = 0; i < HS_ENTRIES; i++)
         {
            if(!(h->entries[i].score)) //when we find a 0 score
               break;
         } //i is now number of entries
         entries = i;

         if(!entries) //if no entries (will always be at least one), break
            break;

         len = 2 + entries*sizeof(HST_ENTRY); //calculate output file length
         buf = ast_alloc(len); //alloc space for it
         buf[0] = entries;     //set up file-byte 0 is number of entries

         memcpy(buf + 1, h->entries, entries*sizeof(HST_ENTRY)); //set up file-remaining bytes (except last byte) are table entries
         buf[len - 1] = CheckSum(buf, len); //set up file-last byte is checksum

         Crypt(buf, len); //encrypt file

         f = OpenHiScoreTable("wb"); //write it
         if(f)
            fwrite(buf, 1, len, f);

      } while(0);

      if(buf)
         free(buf);
      if(f)
         fclose(f);
   }
}

//draws the table
void DrawHiScoreTable(const HISCORE_TABLE * h)
{
   int i;

   if(h)
   {
      glDisable(GL_LIGHTING);
      glColor4f(0.5f, 0.25f, 1.0f, 0.25f);
      glBegin(GL_QUADS);
      glVertex3f(9.0f,  11.5f, 40.0f);
      glVertex3f(31.0f, 11.5f, 40.0f);
      glVertex3f(31.0f, 24.1f, 40.0f);
      glVertex3f(9.0f,  24.1f, 40.0f);
      glEnd();
      glEnable(GL_LIGHTING);

      RenderText(9.31448f, 23.0f, 50.0f, 0.6f, 0.25f, 1.0f, 0.25f, "Player          | Lvl |      Score");
      RenderText(9.31448f, 22.0f, 50.0f, 0.6f, 0.25f, 1.0f, 0.25f, "--------------- | --- | ----------");
      for(i = 0; i < HS_ENTRIES; i++)
      {
         RenderText(9.31448f, 21.0f - (float)i, 50.0f, 0.6f, 0.25f, 1.0f, 0.25f, h->entries[i].name);
         RenderText(19.37144f, 21.0f - (float)i, 50.0f, 0.6f, 0.25f, 1.0f, 0.25f, "| %3lu | %10lu", h->entries[i].level, h->entries[i].score);
      }
   }
}

//returns stdio.h FILE object for hiscore table (close with fclose)
FILE * OpenHiScoreTable(char * mode)
{
#ifdef _WIN32
   return fopen(HS_FILE, mode);
#else
   char szFilename[FILENAME_MAX];
   snprintf(szFilename, FILENAME_MAX - 1, "%s/" HS_FILE, getenv("HOME"));
   szFilename[FILENAME_MAX - 1] = '\0';
   return fopen(szFilename, mode);
#endif
}

//uses simple xor encryption (with a long key) to encrypt a buffer
void Crypt(char * buf, unsigned int len)
{
   //the key is somewhat long, has a decent (though not great) range of values,
   //isn't very easy to figure out (without poking into executable),
   //and changes each time it compiles--it's not a bad key for its purpose
   static const char key[]  = {__TIME__ "\a\b#\x9a\f/" ASTEROIDS_VERSION "\x83\xc9\xaa}\xff\x89`\xeb" __DATE__ "\x01$\"\xb0x\xe3" __FILE__ "\n\a\x99\xb8\r\b\xa3&whee!\">C~\x83"};
   static const int  keylen = sizeof(key)/sizeof(key[0]) - 1;

   unsigned int i;

   if(buf)
   {
      for(i = 0; i < len; i++)
         buf[i] ^= ~(char)((int)key[i % keylen] + 0xa3);
   }
}

//gets the checksum of buffer (minus last byte)
char CheckSum(const char * buf, unsigned int len)
{
   unsigned long sum = 0;
   unsigned int i;
   for(i = 0; i < len - 1; i++)
      sum += buf[i];

   return (char)sum;
}

#ifdef _WIN32
//dialog proc for hiscore dialog
INT_PTR CALLBACK HiScoreDialog(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
   switch(uMsg)
   {
      case WM_INITDIALOG:
         return TRUE;

      case WM_COMMAND:
         switch(LOWORD(wParam))
         {
            case IDOK:
               GetDlgItemText(hWndDlg, IDC_NAME, g_name, HS_NAME_LENGTH);
               //fall through
            case IDCANCEL:
               EndDialog(hWndDlg, (LOWORD(wParam) == IDOK));
               return TRUE;
         }
         break;
   }

   return FALSE;
}
#endif
